/*
=============================================================
proxy_data.c
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the co-operating proxy program. The co-operating proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover power management services will make it possible for devices with 
these services to go into power sleep without disappearing from the network by 
sending out spoofed HTTP OK and SSDP:alive. When there is an incoming TCP 
connection for the sleeping device the proxy will wake it up using WOL.

This file contains functions used for the data structures in this program.
The data structures are of type netwib_ring (linked lists) and the functions
in this file are used to find, extract and delete items from the structure.

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program at your own risk.
=============================================================
*/

#include <stdio.h>
#include <string.h>
#include "proxy_funcs.h"

//finds a device with the specified ip
//input: *ip (the address to look for), *list (the list to look in)
//output: ptr (the element found)
//returns NETWIB_ERR_NOTFOUND if nothing is found
netwib_err find_ip(netwib_ip *ip, netwib_ring *list, netwib_ptr *pptr){
//--------------declaration-------------------------------
  netwib_ring_index *pringindex;
  netwib_ptr item;
  upnp_device *pdevice;
  netwib_err ret;
  netwib_cmp cmp;
//--------------end of declaration------------------------
      
  ret = netwib_ring_index_init(list, &pringindex);
  if(ret != NETWIB_ERR_OK){
    netwib_fmt_display("Error while initiating index in find_ip.");
    return(NETWIB_ERR_NOTFOUND);
  }
  cmp = NETWIB_CMP_LT;
      
  while(cmp != NETWIB_CMP_EQ){
    //get the next device            
    ret = netwib_ring_index_next(pringindex, &item);
    if(ret == NETWIB_ERR_OK){                   
      pdevice = (upnp_device *)item;                    
      //check the ip
      netwib_ip_cmp(ip, &pdevice->ip, &cmp);                           
    }
    else{//no more devices in the list
      netwib_ring_index_close(&pringindex);
      return(NETWIB_ERR_NOTFOUND);
    }
  }
  netwib_ring_index_close(&pringindex);          
  *pptr = (netwib_ptr *)pdevice;
  return(NETWIB_ERR_OK);              
}


//finds a service with the specified usn
//input: usn (the usn to look for), *list (the list to look in)
//output: ptr (the element found)
//returns NETWIB_ERR_NOTFOUND if nothing is found
netwib_err find_usn(netwib_string usn, netwib_ring *list, netwib_ptr *pptr){
//--------------declaration-------------------------------
  netwib_ring_index *pringindex;
  netwib_ptr item;
  service *pservice;
  netwib_err ret;
  int i;
//--------------end of declaration------------------------
      
  ret = NETWIB_ERR_OK;
  ret = netwib_ring_index_init(list, &pringindex);
  if(ret != NETWIB_ERR_OK){
    netwib_fmt_display("Error while initiating index in find_usn.");
    return(NETWIB_ERR_NOTFOUND);
  }
  i = 1;
  while(i != 0){
    //get the next device            
    ret = netwib_ring_index_next(pringindex, &item);
    if(ret == NETWIB_ERR_OK){                   
      pservice = (service *)item;
      //check the usn of the service
      i = netwib_c_strcmp(usn, pservice->usn);                           
    }
    else{//no more devices in the list, no matching service found
      netwib_ring_index_close(&pringindex);
      return(NETWIB_ERR_NOTFOUND);
    }
  }
  netwib_ring_index_close(&pringindex);          
  *pptr = (netwib_ptr *)pservice;
  return(NETWIB_ERR_OK);            
}

//frees memory when a service is deleted from a ring
//input: item (void pointer to the item)
netwib_err delete_service_from_list(netwib_ptr item){
//--------------declaration-------------------------------
  netwib_err ret;
//--------------end of declaration------------------------
  ret = netwib_ptr_free(&item);        
  return ret;         
}

//frees memory when a upnp_device is deleted from a ring
//input: item (void pointer to the item)
netwib_err delete_device_from_list(netwib_ptr item){
//--------------declaration-------------------------------
  upnp_device *dev;
  netwib_err ret;
//--------------end of declaration------------------------
  dev = (upnp_device *)item;
  netwib_ring_close(&dev->services, NETWIB_TRUE);
  ret = netwib_ptr_free(&item);      
  return ret;         
}

//compare an item in a list with an ip
//used for proxy_devices and upnp_devices
//input: item (item to compare), pinfos (ip to compare with)
//output: *b (result of the comparison)
netwib_err compare_ip(netwib_constptr item, netwib_ptr pinfos, netwib_bool *b ){
//--------------declaration-------------------------------
  upnp_device *dev;
  netwib_ip *pip;
  netwib_cmp cmp;
//--------------end of declaration------------------------
  dev = (upnp_device *)item;  
  pip = (netwib_ip *)pinfos;
  netwib_ip_cmp(pip, &dev->ip, &cmp);
  if(cmp == NETWIB_CMP_EQ){
    *b = NETWIB_TRUE;
  }
  else{
    *b = NETWIB_FALSE;
  }
  return NETWIB_ERR_OK;       
}

//remove the device with ip *ip if it is found
//input: *ip (address of device to remove), *ring (ring to remove from)
netwib_err remove_device(netwib_ip *ip, netwib_ring *ring){
//--------------declaration-------------------------------
  netwib_err ret;
  netwib_ptr ptr;
//--------------end of declaration------------------------
  ret = find_ip(ip, ring, &ptr);
  if(ret == NETWIB_ERR_OK){//device found, delete it
    ret = netwib_ring_del_criteria(ring, &compare_ip, (netwib_ptr *)ip, NETWIB_TRUE);
    return ret;
  }
  else { //device not found
    return NETWIB_ERR_NOTFOUND;
  }
}

//find the i:th time hdr appears in a xml packet
//input: *rest(the packet to parse), hdr, i
//output pstr containing what is between hdr and next < (end of header)
netwib_err parse_xml(netwib_buf *rest, netwib_string hdr, char *pstr, int i ){
//--------------declaration-------------------------------
  netwib_string bufstring;
  netwib_buf tempbuf;
  char *pc;
  netwib_err ret;
  int j,k;
  netwib_uint64 shift;
  char teststr[100];
//--------------end of declaration------------------------        
  ret = NETWIB_ERR_OK;
  //copy buffer to a string
  teststr[0] = '<';
  teststr[1] = '\0';
  netwib_c_strcat(teststr, hdr);
  netwib_c_strcat(teststr, ">\0");
  netwib_buf_init_mallocdefault(&tempbuf);
  netwib_buf_append_buf(rest, &tempbuf);
  ret =  netwib_buf_ref_string(&tempbuf, &bufstring);
  if(ret != NETWIB_ERR_OK) {
	netwib_fmt_display("Error while initiating search string in parse_xml\n");
	return(NETWIB_ERR_NOTFOUND);
  }
     
  //find hdr
  for(k=1; k<=i; ++k){
    pc = netwib_c_strcasestr(bufstring, teststr);
    if(pc == NULL) { //the string wasn't found
  	  pstr = NULL;
	  return(NETWIB_ERR_NOTFOUND);
    }
    //need to shift away what is before hdr to be able to find the next appearens
	shift = pc - bufstring; 
	if(shift == 0) { //this means we found searchstr in the begining. Need to shift this away otherwise we will find the same place all the time
  	  netwib_buf_shift(&tempbuf,	-netwib_c_strlen(teststr), -netwib_c_strlen(teststr));
  	  --k;
    }
    else {
	  netwib_buf_shift(&tempbuf, -shift, -shift);
	  ret =  netwib_buf_ref_string(&tempbuf, &bufstring);
      if(ret != NETWIB_ERR_OK) {
	 	netwib_fmt_display("Error while re-initiating search string in parse_xml\n");
	 	return(NETWIB_ERR_NOTFOUND);
	  }
	}
  }   
  pc = bufstring + netwib_c_strlen(hdr)+2; //skip header and :
       
  //skip space
  if(*pc == ' ')
    pc++;
    
  j = 0;
  while((pstr[j] = *pc) != '<'){
    ++j;
    pc++;         
  }
  pstr[j] = '\0';
  //netwib_fmt_display("pstr = %s\n", pstr);
  return(NETWIB_ERR_OK);
}

//gets port and url from a string containing ip:port/url
//input: bufstring(containing the string), *pip(ip of the string)
//output: *port, *url
netwib_err get_port_and_url(netwib_string location, char *port, char *url, netwib_constip *dstip){
//--------------declaration-------------------------------		   
  char *pc;
  int i;
  netwib_buf tempbuf;
  netwib_string bufstring;
//--------------end of declaration------------------------
  netwib_buf_init_mallocdefault(&tempbuf);
  netwib_buf_append_ip(dstip, NETWIB_IP_ENCODETYPE_IP, &tempbuf);
  netwib_buf_ref_string(&tempbuf, &bufstring);
  pc = netwib_c_strcasestr(location, bufstring);
  pc = pc+netwib_c_strlen(bufstring)+1;
  i = 0;
  while((port[i] = *pc) != '/'){
    ++i;
    pc++;         
  }
  port[i] = '\0';
  i = 0;
  while((url[i] = *pc) != '\0'){
    ++i;
    pc++;         
  }
  url[i] = '\0';  		   
}
