/*
=============================================================
proxy_funcs.h
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the invisible proxy program. The invisible proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover sleeping devices and make sure they still seem to be connected
to the network by spoofing HTTP OK and SSDP:alive. When there is
an incoming TCP connection for the device the proxy will wake it up using
WOL.

This is the definition file that defines some structures and 
functions used in the proxy program.

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program on your own risk.
=============================================================
*/

#ifndef PROXY_FUNCS
#define PROXY_FUNCS

#include <netwib.h>

//struct used to send input to threads
typedef struct thread_input thread_input;
struct thread_input{
       netwib_ips *ips;
       netwib_ring *proxy;
       netwib_ring *cache;
       netwib_ip *myip;
       netwib_port myport;
       
       };

//defines an UPnP device
//contains all services located on the same IP
typedef struct upnp_device upnp_device;

struct upnp_device{
  netwib_ip ip;
  netwib_eth eth;
  netwib_ring *services; //contains all embedded services
  char server[100];
  char location[200];
  netwib_time last_activity;
  //netwib_bool proxy;
};

//defines a service
typedef struct service service;
struct service{
  char st[100];
  char usn[200];
  netwib_time last_notify;
  netwib_uint16 timeout;
  netwib_port port;
};

//will parse the packet in *in to the headers
//input: in, dlt
//output: plinkhdr, piphdr, pudphdr, rest
netwib_err parse_packet(netwib_buf *in, netwib_device_dlttype dlt, 
                       netwib_linkhdr *plinkhdr, netwib_iphdr *piphdr,
                       netwib_udphdr *pudphdr, netwib_buf *rest);

//will sniff the network 
//sniffed packet will be returned
//input: device, filter, wait (time to sniff), pio (input channel)
//output: iphdr, udphdr, ethernetaddresses and rest of the sniffed packet
netwib_err sniff(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_buf *rest,
                               netwib_time *wait, netwib_io *pio);
							   
//will multicast a SSDP:discover
//input: 
//output:
netwib_err send_discovery(netwib_port myport);

//find a header in a SSDP packet and retrieve it
//used for getting NTS, ST, USN etc from HTTP
//input: *rest (the buffer that contains the whole HTTP message), *phdr (header to locate)
//output: *pstr (must be vector)
netwib_err get_hdr(netwib_buf *rest, char* phdr, char *pstr );

//sends a WOL packet
//input: *eth (the ethernet address to send to)
netwib_err send_wol(netwib_eth *eth);

//thread function to udate which devices to proxy
netwib_err update_cache(netwib_ptr infosin, netwib_ptr *pinfosout);

//thread function to spoof SSDP-notifies when needed
netwib_err spoof_notify(netwib_ptr infosin, netwib_ptr *pinfosout);

//add a device to the proxy list
netwib_err add_to_proxy(upnp_device *dev);

//remove device from proxy list
netwib_err remove_from_proxy(upnp_device *dev);

//finds a device with the specified ip
//input: *ip (the address to look for), *list (the list to look in)
//output: ptr (the element found)
//returns NETWIB_ERR_NOTFOUND if nothing is found
netwib_err find_ip(netwib_ip *ip, netwib_ring *list, netwib_ptr *pptr);

//finds a service with the specified usn
//input: usn (the usn to look for), *list (the list to look in)
//output: ptr (the element found)
//returns NETWIB_ERR_NOTFOUND if nothing is found
netwib_err find_usn(netwib_string usn, netwib_ring *list, netwib_ptr *pptr);

//calculates checksum for an iphdr
netwib_err ip_checksum(netwib_iphdr *iphdr);

//sends a spoofed message in the buffert buf on the device device
netwib_err spoof_send(netwib_buf *buf, netwib_buf *device);

//Sends a SSDP:alive for the service serv
//input: *serv, the service for which to send a SSDP:alive
netwib_err send_alive(service *serv, upnp_device *dev);

//Sends a HTTP OK
//input: *serv (the service to send from), *dev (the device to send from)
//*send device (the initiated nic), to_ip (ip to send to), to_eth (ethernet to send to)
//to_port (port to send to)
netwib_err send_httpok(service *serv, upnp_device *dev, netwib_buf *send_device, 
           netwib_ip *to_ip, netwib_eth *to_eth, netwib_port to_port );
           
//find the timeout (Cache-control max-age) and parse it to an int value
//input: *rest (packet)
//output: *timeout (the int value in seconds of the timeout)
netwib_err get_timeout(netwib_buf *rest, netwib_uint16 *timeout);

//adds the device or service to the ring,
//if it is already there the timestamp is updated
//only adds new device if it is a Powermgmt service, then the XML scheme is fetched
//input: *iphdr, *ethhdr, *udphdr, myport(port of this proxy), eventport (port to receive events on), 
//myip, *rest (the packet),
//*upnp_devices (linked list to add to)
netwib_err process_ssdp_alive(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_port myport, netwib_port eventport,
                               netwib_ip *myip, netwib_buf *rest,
                               netwib_ring *upnp_devices);
                               
//frees memory when a service is deleted from a ring
//input: item (void pointer to the item)
netwib_err delete_service_from_list(netwib_ptr item);

//frees memory when a upnp_device is deleted from a ring
//input: item (void pointer to the item)
netwib_err delete_device_from_list(netwib_ptr item);

//compare two items in a list
//used for proxy_devices and upnp_devices
//input: item1, item2 (items to compare)
//output: *b (result of the comparison)
netwib_err compare_ip(netwib_constptr item, netwib_ptr pinfos, netwib_bool *b);

//remove the device with ip *ip if it is found
//input: *ip (address of device to remove), *ring (ring to remove from)
netwib_err remove_device(netwib_ip *ip, netwib_ring *ring);

//get an arp packet from the io
//input: *pio
//output *arp
netwib_err get_arp(netwib_io *pio, netwib_arphdr *arp);

//send arp packet
//input: *from_ip, *to_ip, *from_eth, *to_eth, *send_device
netwib_err send_arp(netwib_ip *from_ip, netwib_ip *to_ip, netwib_eth *from_eth, netwib_eth *to_eth,
                                        netwib_buf *send_device);

//Initiate a NIC
//input: hwt (hardware type, which kind of device is to be initiated)
//output: *pbuf (will contain description of the device), 
netwib_err init_device(netwib_buf *pbuf, netwib_device_hwtype hwt);

//Thread that reads from the event socket and updates cache and proxy
//cache if an event is received, reads the socket every second
netwib_err read_event_socket(netwib_ptr infosin, netwib_ptr *pinfosout);

//Process an incoming HTTP OK packet
//input: *iphdr, *ethhdr, *udphdr, *rest, myport, *ti
netwib_err process_HTTPOK_packet(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_buf *rest, netwib_port myport, netwib_ring *upnp_devices, 
							   netwib_port eventport, netwib_ip *myip);

//Get xml for a device that has Powermgmnt service and subscribes to the Powermgmnt service
//input: *ti, thread info with myip, eventport and the two cache rings, *dstip, *deveth, 
//        myport, *rest containing the incoming HTTP OK or SSDP alive
netwib_err get_xml_and_subscribe(netwib_constip *dstip, netwib_consteth *deveth, netwib_port myport,
                                netwib_buf *rest, netwib_ring *upnp_devices, netwib_port eventport, netwib_constip *myip);




//find the i:th time hdr appears in a xml packet, parse out string between hdr and next < (end of header)
//input: *rest, containing the packet to parse, *hdr, the header to find, i
//output *pstr                
netwib_err parse_xml(netwib_buf *rest, netwib_string parse, char *pstr, int i);

//gets port and url from a string containing ip:port/url
//input: bufstring, *pip
//output: *port, *url
netwib_err get_port_and_url(netwib_string bufstring, char *port, char *url, netwib_constip *pip);




#endif
