/*
=============================================================
proxy_receive.c
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the co-operating proxy program. The co-operating proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover power management services will make it possible for devices with 
these services to go into power sleep without disappearing from the network by 
sending out spoofed HTTP OK and SSDP:alive. When there is an incoming TCP 
connection for the sleeping device the proxy will wake it up using WOL.

This file contains functions for receiving and parsing packets.
There are functions for retrieveng SSDP headers from raw packets
and a function for sniffing packets. Parsed packets are processed
and services and devices are added to the data structure.

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program at your own risk.
=============================================================
*/

#include <stdio.h>
#include <string.h>
#include "proxy_funcs.h"


//will parse the packet in *in to the headers
//input: in, dlt
//output: plinkhdr, piphdr, pudphdr, rest
netwib_err parse_packet(netwib_buf *in, netwib_device_dlttype dlt, 
                       netwib_linkhdr *plinkhdr, netwib_iphdr *piphdr,
                       netwib_udphdr *pudphdr, netwib_buf *rest) 
{ 
//--------------declaration---------------------------
  netwib_err ret;
  netwib_uint32 skipsize;
//--------------end of declaration------------------------
      
  netwib_buf_append_buf(in, rest);
  ret = netwib_pkt_decode_linkhdr(dlt , rest, plinkhdr, &skipsize); // get the linkhdr
  if(ret != NETWIB_ERR_OK){
	netwib_er(netwib_fmt_display("Error while when decoding linkheader."));
	return(ret);
  }
  netwib_er(netwib_buf_shift(rest, -skipsize, 0)); // shift away linkhdr
  if(plinkhdr->hdr.ether.type == NETWIB_ETHERHDRTYPE_ARP){
    //we have an arp, stop decoding
    return NETWIB_ERR_NOTCONVERTED;
  }
  ret = netwib_pkt_decode_iphdr(rest, piphdr, &skipsize); //get the iphdr
  if(ret != NETWIB_ERR_OK){
    netwib_er(netwib_fmt_display("Error while when decoding ipheader."));
	return(ret);
  }
  if(piphdr->protocol == NETWIB_IPPROTO_TCP) {
    return(NETWIB_ERR_DATAOTHERTYPE); //we got a tcp packet, stop decoding
  }

  netwib_er(netwib_buf_shift(rest, -skipsize, 0)); // shift away iphdr
  ret = netwib_pkt_decode_udphdr(rest, pudphdr, &skipsize); //get the udphdr
  if(ret != NETWIB_ERR_OK){
	netwib_er(netwib_fmt_display("Error while when decoding udpheader."));
	return(ret);
  }
  netwib_er(netwib_buf_shift(rest, -skipsize, 0)); //shift away udphdr
  return(ret);                               
}


//will sniff the network 
//sniffed packet will be returned
//input: device, filter, wait (time to sniff), pio (input channel)
//output: iphdr, udphdr, ethernetaddresses and rest of the sniffed packet
netwib_err sniff(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_buf *rest,
                               netwib_time *wait, netwib_io *pio)
{
//--------------declaration---------------------------
  netwib_buf packet;
  netwib_device_dlttype dlt;
  netwib_err ret;
  netwib_linkhdr linkhdr;
  netwib_bool event;
  char usn[100], st[100];
//--------------end of declaration------------------------

  ret = NETWIB_ERR_OK;
  netwib_er(netwib_buf_init_mallocdefault(&packet));    
  netwib_er(netwib_sniff_ctl_get_dlt(pio, &dlt)); // get dlt
  netwib_er(netwib_io_wait_read(pio, wait, &event)); // wait for data in io

  //does not always give an event when there is something to read
  //read from the io, get the packet
  ret = netwib_io_read(pio, &packet);
  if(ret == NETWIB_ERR_NOTCONVERTED) {
	netwib_er(netwib_fmt_display("Error while receiving: Data not converted"));		 
	netwib_buf_close(&packet);
	return(NETWIB_ERR_NOTFOUND);
  }
  else if(ret == NETWIB_ERR_DATAMISSING) {
	netwib_er(netwib_fmt_display("Error while recieving: Datamissing"));
	netwib_buf_close(&packet);
	return(NETWIB_ERR_NOTFOUND);
  }
  else if (ret == NETWIB_ERR_DATANOTAVAIL) {
	netwib_er(netwib_fmt_display("."));
	netwib_buf_close(&packet);
	return(NETWIB_ERR_NOTFOUND);
  }
  else if (ret == NETWIB_ERR_OK) { //we actually received something 
	//parse the received packet into headers
	ret = parse_packet(&packet, dlt, &linkhdr, iphdr, udphdr, rest);     
	if(ret != NETWIB_ERR_OK && ret!= NETWIB_ERR_DATAOTHERTYPE ){
	  //this has to be an arp
	  //put data back to the io and exit
	  netwib_io_unread(pio, &packet);
	  netwib_buf_close(&packet); 
	  return NETWIB_ERR_NOTCONVERTED;		
	}
  }
  *ethhdr = linkhdr.hdr.ether; //only need the ethernet header from the linkheader
  //close open buffers and exit	
  netwib_buf_close(&packet); 
  return(ret);		
}

//find a header in a SSDP packet and retrieve it
//used for getting NTS, ST, USN etc from HTTP
//input: *rest (the buffer that contains the whole HTTP message), hdr (header to locate)
//output: *pstr (must be vector)
netwib_err get_hdr(netwib_buf *rest, netwib_string hdr, char *pstr ){
//--------------declaration-------------------------------
  netwib_string bufstring;
  char *pc;
  netwib_err ret;
  int i;
//--------------end of declaration------------------------        

  ret = NETWIB_ERR_OK;
  //copy buffer to a string
  netwib_er(netwib_buf_ref_string(rest, &bufstring));
  //find hdr
  pc = netwib_c_strcasestr(bufstring, hdr);    
  pc = pc+netwib_c_strlen(hdr); //skip header and :
  //skip space
  if(*pc == ' ')
    pc++;
  i = 0;
  while((pstr[i] = *pc) != '\r'){
    ++i;
    pc++;         
  }
  pstr[i] = '\0';   
}

//find the timeout (Cache-control max-age) and parse it to an int value
//input: *rest (packet)
//output: *timeout (the int value in seconds of the timeout)
netwib_err get_timeout(netwib_buf *rest, netwib_uint16 *timeout){
//--------------declaration-------------------------------
  netwib_string bufstring;
  char *pc;
  netwib_err ret;
  netwib_uint16 tmp;
  int i;
//--------------end of declaration------------------------ 
  
  ret = NETWIB_ERR_OK;
  //copy buffer to a string
  netwib_er(netwib_buf_ref_string(rest, &bufstring));
  //find = that is before numerical value
  pc = netwib_c_strcasestr(bufstring, "=");     
  pc++; //skip =
  //skip space
  if(*pc == ' ')
    pc++;
  tmp = 0;
  //convert string to int
  for(; *pc >='0' && *pc <='9'; pc++)
    tmp = 10*tmp+(*pc - '0');
  *timeout = tmp;
  return (ret);    
}

//adds the device or service to the device cache,
//if it is already there the timestamp is updated
//only adds new device if it is a Powermgmt service, then the XML scheme is fetched
//input: *iphdr, *ethhdr, *udphdr, myport(port of this proxy), eventport (port to receive events on), 
//myip, *rest (the packet),
//*upnp_devices (linked list to add to)
netwib_err process_ssdp_alive(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_port myport, netwib_port eventport,
                               netwib_ip *myip, netwib_buf *rest,
                               netwib_ring *upnp_devices){
//--------------declaration-------------------------------
  upnp_device *upnpdev;
  service *upnp_service;
  char str[100];
  netwib_ptr ptr;
  netwib_err ret;
  thread_input ti;
//--------------end of declaration------------------------ 

  //search for device on IP
  ret = find_ip(&iphdr->src, upnp_devices, &ptr);                    		       
  if(ret == NETWIB_ERR_OK){ //device is in list
    upnpdev = (upnp_device *)ptr;
    get_hdr(rest, "USN:", str);
    ret =find_usn(str, upnpdev->services, &ptr);                     
    if(ret == NETWIB_ERR_NOTFOUND){
      //and service is not in list, add the service
      //not likely to happen with power management
      netwib_er(netwib_ptr_malloc(sizeof(service), &ptr));                              
 	  upnp_service = (service *) ptr;
      get_hdr(rest, "\r\nNT:", upnp_service->st);                   
      get_hdr(rest, "USN:", upnp_service->usn);
 	  get_timeout(rest, &upnp_service->timeout);	          
 	  upnp_service->port = udphdr->src;
 	  netwib_er(netwib_time_init_now(&upnp_service->last_notify));
 	  ptr = (netwib_ptr *)upnp_service;
 	  netwib_ring_add_first(upnpdev->services, ptr); 
      netwib_fmt_display("Added service to device %{ip}\n", &upnpdev->ip);      
    }
    else {//if device in list and service in list
       //and if packet is SSDP:alive, update timestamp
       upnp_service = (service *) ptr;
       netwib_er(netwib_time_init_now(&upnp_service->last_notify));
       netwib_er(netwib_time_init_now(&upnpdev->last_activity));
    }                               	         	
  }	         	
  else {
    //Device is not in list, if it has power mgmt service
    //get the xml
    //create a new device
    get_hdr(rest, "\r\nNT:", str);
    if(strncmp(str, "urn:schemas-power:service:Powermgmt",35)==0){
      //netwib_fmt_display("New power mgmt service found, adding to cache\n");                        
      get_xml_and_subscribe(&iphdr->src, &ethhdr->src, myport,
                             rest, upnp_devices, eventport, myip);
    }
  }                                           
}


//get an arp packet from the io
//input: *pio containing the arp packet
//output *arp
netwib_err get_arp(netwib_io *pio, netwib_arphdr *arp){
//--------------declaration-------------------------------
  netwib_err ret;
  netwib_buf packet;
  netwib_uint32 skipsize;
//--------------end of declaration------------------------
  netwib_er(netwib_buf_init_mallocdefault(&packet));
  ret = netwib_io_read(pio, &packet);
  if(ret == NETWIB_ERR_NOTCONVERTED) {
	netwib_er(netwib_fmt_display("Error while receiving: Data not converted"));		 
	netwib_buf_close(&packet);
	return(NETWIB_ERR_NOTFOUND);
  }
  else if(ret == NETWIB_ERR_DATAMISSING) {
    netwib_er(netwib_fmt_display("Error while recieving: Datamissing"));
	netwib_buf_close(&packet);
	return(NETWIB_ERR_NOTFOUND);
  }
  else if (ret == NETWIB_ERR_DATANOTAVAIL) {
    netwib_er(netwib_fmt_display("IO empty.\n"));
	netwib_buf_close(&packet);
	return(NETWIB_ERR_NOTFOUND);
  }
  else if (ret == NETWIB_ERR_OK) {
	//parse the received packet into headers
	netwib_er(netwib_buf_shift(&packet, -14, 0)); // shift away linkhdr
	netwib_er(netwib_pkt_decode_arphdr(&packet, arp, &skipsize));
	return NETWIB_ERR_OK;       
  }       
}

//Thread that reads from the event socket and updates cache and proxy
//cache if an event is received, reads the socket every second
netwib_err read_event_socket(netwib_ptr infosin, netwib_ptr *pinfosout){
//--------------declaration-------------------------------
  netwib_err ret, ret2;
  netwib_buf packet, ip, http_ok;
  char host[100], value[50];
  char *pc;
  thread_input *pti;
  netwib_ptr ptr;
  netwib_port remoteport;
  netwib_ip remoteip;
  netwib_io *pio;
  netwib_bool in_ips;
//--------------end of declaration------------------------

  pti = (thread_input *)infosin;
  netwib_buf_init_ext_text("HTTP/1.1 200 OK\r\n", &http_ok);
  netwib_er(netwib_buf_init_mallocdefault(&packet)); 
  while(NETWIB_TRUE){
    ret = netwib_io_init_sock_tcp_ser(pti->myip, pti->myport, &pio);
    if(ret != NETWIB_ERR_OK){
      netwib_er(netwib_fmt_display("Error opening event socket in thread\n"));
    } 
        
    //read from the socket, waits until there is data in the socket
    ret = NETWIB_ERR_NOTFOUND;
    while(ret != NETWIB_ERR_OK){
      netwib__buf_reinit(&packet);
      ret = netwib_io_read(pio, &packet); 
      netwib_fmt_display("*");
    }
                  
    netwib_er(netwib_sock_ctl_get_remote(pio, &remoteip, &remoteport));
    netwib_fmt_display("Received event from %{ip}\n", &remoteip);     	
    //check if we the device is in the cache
    //otherwise we're not interested in the message
    ret2 = find_ip(&remoteip, pti->cache, &ptr);
    netwib_ips_contains_ip(pti->ips, &remoteip, &in_ips);
    if ((ret ==NETWIB_ERR_OK) && (ret2 == NETWIB_ERR_OK)) {                   
      ret =parse_xml(&packet, "power", value, 1);
      if(ret == NETWIB_ERR_OK && value[0]=='f' && !in_ips){
        netwib_fmt_display("%{ip} is powering down, start proxying\n", &remoteip);
        //device is powering down, start proxying
        netwib_ips_add_ip(pti->ips, &remoteip);
        copy_from_cache_to_proxy(&remoteip, pti->cache, pti->proxy);
      }
      else if (ret == NETWIB_ERR_OK && value[0]=='t' && in_ips){
        //device is powering up, remove from proxy
        netwib_fmt_display("%{ip} is powering up, remove from proxy\n",&remoteip );
        netwib_ips_del_ip(pti->ips, &remoteip);
        remove_device(&remoteip, pti->proxy);
      }
      else if (ret != NETWIB_ERR_OK){
        netwib_er(netwib_fmt_display("Error when parsing XML in read_event_socket\n"));
      }
    }
    //send a HTTP ok back
    netwib_io_write(pio, &http_ok); 
    netwib_io_close(&pio);            
  }   
  netwib_buf_close(&ip);
  netwib_buf_close(&packet); 
  netwib_buf_close(&http_ok); 
  return(NETWIB_ERR_OK);        
}

//checks a HTTP OK packet
//checks if it is from a Powermgmnt service. Will send out discovery for device if needed
//input: *iphdr, *ethhdr, *udphdr, *rest, myport, *ti
netwib_err process_HTTPOK_packet(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_buf *rest, netwib_port myport, netwib_ring *upnp_devices, 
							   netwib_port eventport, netwib_ip *myip){
//--------------declaration-------------------------------
  netwib_string bufstring;
  netwib_ptr ptr;
//--------------end of declaration------------------------ 
  
  netwib_er(netwib_buf_ref_string(rest, &bufstring));
  if(netwib_c_strcasestr(bufstring, "urn:schemas-power:service:Powermgmt") == NULL) { //not from a powermgmnt service process as normal
    netwib_fmt_display("Error. HTTP OK received is  not a powermgmnt HTTP-OK\n");
	return(NETWIB_ERR_NOTFOUND);	  
  }
  else {
	if(find_ip(&iphdr->src, upnp_devices, &ptr) == NETWIB_ERR_NOTFOUND)
  	  get_xml_and_subscribe(&iphdr->src, &ethhdr->src, myport, rest, upnp_devices, eventport, myip);
    else
      netwib_fmt_display("Device already in cache");
  }
  return(NETWIB_ERR_OK);  
}
