/*
=============================================================
proxy_send.c
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the co-operating proxy program. The co-operating proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover power management services will make it possible for devices with 
these services to go into power sleep without disappearing from the network by 
sending out spoofed HTTP OK and SSDP:alive. When there is an incoming TCP 
connection for the sleeping device the proxy will wake it up using WOL.

This file contains functions for sending and spoofing messages such
as SDDP:discover, SSDP:alive, HTTP OK, ARP and WOL.

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program at your own risk.
=============================================================
*/

#include <stdio.h>
#include <netwib.h>
#include "proxy_funcs.h"

//will multicast a SSDP:discover for powermanagement services
//input: myport (port to send from)
//output: none
netwib_err send_discovery(netwib_port myport)
{
//--------------declaration-------------------------------
  netwib_err ret;
  netwib_buf msearch_packet, tempbuf;
  netwib_ip ip_multicast;
  netwib_port port_multicast;
  netwib_io *pio;
  netwib_uint32 ttl = 4; //should not be more than 4 according to UPnP Forum standard
//--------------end of declaration------------------------

  netwib_er(netwib_ip_init_ip4_fields(239,255,255,250, &ip_multicast));
  port_multicast = 1900;
  ret = NETWIB_ERR_OK;
  netwib_er(netwib_buf_init_mallocdefault(&msearch_packet));
   
  //open a multicast UDP socket, has only TTL = 1 !!!!!
  ret = netwib_io_init_sock_udp_cli(NULL, &ip_multicast, myport, port_multicast, &pio);
  if(ret != NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Error when opening socket in send_discovery.\n"));
    return(ret);
  }
  //change the TTL
  ret = netwib_sock_ctl_set_multicastttl(pio,ttl);
  if(ret != NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Error when changing TTL send_discovery.\n"));
  }

  //build the M-search packet
  netwib_buf_init_ext_text("M-SEARCH * HTTP/1.1\r\nMX: 10\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nST: urn:schemas-power:service:Powermgmt\r\n\r\n", &tempbuf); //ssdp:all\r\n\r\n", &tempbuf);
  netwib_buf_decode(&tempbuf, NETWIB_DECODETYPE_DATA, &msearch_packet);
  //send the packet twice
  ret = netwib_io_write(pio, &msearch_packet);
  ret = netwib_io_write(pio, &msearch_packet);
  if(ret != NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Error when writing to socket in send_discovery\n"));
    return(ret);
  }
  netwib_er(netwib_buf_close(&msearch_packet));
  netwib_er(netwib_io_close(&pio));
  return(ret);         
}


//sends a WOL packet
//input: *eth (the ethernet address to send to)
netwib_err send_wol(netwib_eth *ethr){
//--------------declaration-------------------------------  
  netwib_buf pkt, tmpbuf;
  netwib_err ret;
  netwib_ip ip_broadcast;
  netwib_port port;
  netwib_io *pio;
  netwib_eth eth_broadcast;
  int i, j;
//--------------end of declaration------------------------        
    
  netwib_er(netwib_ip_init_ip4_fields(255, 255,255,255, &ip_broadcast));
  port = 2536;
  ret = NETWIB_ERR_OK;
  netwib_er(netwib_buf_init_mallocdefault(&pkt));
  netwib_er(netwib_buf_init_mallocdefault(&tmpbuf));
   
  //open a broadcast UDP socket, has only TTL = 1 !!!!!
  ret = netwib_io_init_sock_udp_cli_easy(&ip_broadcast, port, &pio);
  if(ret != NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Error when opening socket in send_wol.\n"));
    return(ret);
  }
  //build the WOL packet
  netwib_eth_init_fields(0xff, 0xff, 0xff, 0xff, 0xff, 0xff, &eth_broadcast);
  for(i = 0; i< NETWIB_ETH_LEN; ++i){
    netwib_buf_append_byte(eth_broadcast.b[i], &tmpbuf);
  }
  for(i = 0; i <17; ++i){
    for(j = 0; j< NETWIB_ETH_LEN; ++j){
      netwib_buf_append_byte(ethr->b[j], &tmpbuf);
    }
  }
  ret = netwib_io_write(pio, &tmpbuf);
  if(ret != NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Error when writing to socket in send_discovery\n"));
    return(ret);
  }
  netwib_er(netwib_buf_close(&pkt));
  netwib_er(netwib_buf_close(&tmpbuf));
  netwib_er(netwib_io_close(&pio));
  netwib_fmt_display("WOL sent to %{eth}\n", ethr);
  return(ret);
}

//Sends a SSDP:alive for the service serv
//input: *serv, the service for which to send a SSDP:alive
netwib_err send_alive(service *serv, upnp_device *dev)
{
//--------------declaration-------------------------------  
  netwib_err ret;
  netwib_time t;
  netwib_buf send_device, tempbuf, notify_packet, sendbuf, timebuf;
  netwib_iphdr iphdr;
  netwib_linkhdr linkhdr;
  netwib_udphdr udphdr;
  netwib_io *pio;
//--------------end of declaration------------------------

  //initiating senddevice, ethhdr, iphdr and udphdr
  netwib_er(netwib_buf_init_mallocdefault(&send_device));
  init_device(&send_device, NETWIB_DEVICE_HWTYPE_ETHER); // initiate sending device
  netwib_linkhdr_initdefault(NETWIB_DEVICE_DLTTYPE_EN10MB, &linkhdr);
  netwib_eth_init_fields(0x01,0x00,0x5e,0x7f,0xff,0xfa,&linkhdr.hdr.ether.dst);
  linkhdr.hdr.ether.src = dev->eth; // set src mac address
  linkhdr.hdr.ether.type = NETWIB_ETHERHDRTYPE_IP4;
  netwib_er(netwib_iphdr_initdefault(NETWIB_IPTYPE_IP4, &iphdr));
  netwib_er(netwib_ip_init_ip4_fields(239,255,255,250,&iphdr.dst));
  iphdr.src = dev->ip;
  iphdr.header.ip4.id = 123;
  iphdr.protocol = NETWIB_IPPROTO_UDP;
  netwib_er(netwib_udphdr_initdefault(&udphdr));
  udphdr.dst = 1900;
  udphdr.src = 1900;
    
  //build the SSDP:alive packet
  netwib_buf_init_mallocdefault(&tempbuf);
  netwib_buf_append_text("NOTIFY * HTTP/1.1\r\nLOCATION: ", &tempbuf);
  netwib_buf_append_text(dev->location, &tempbuf);
  netwib_buf_append_text("\r\nHOST: 239.255.255.250:1900\r\nSERVER: ", &tempbuf);
  netwib_buf_append_text(dev->server, &tempbuf);								  
  netwib_buf_append_fmt(&tempbuf, "\r\nNTS: ssdp:alive\r\nUSN: ", &tempbuf);
  netwib_buf_append_text(serv->usn, &tempbuf);
  netwib_buf_append_fmt(&tempbuf, "\r\nCACHE-CONTROL: max-age=%{uint16}\r\nNT: ", serv->timeout); 
  netwib_buf_append_text(serv->st, &tempbuf);
  netwib_buf_append_text("\r\n\r\n", &tempbuf);
  //netwib_buf_append_text("\r\nContent-Length: 0\r\n\r\n", &tempbuf);//probably not needed
  netwib_er(netwib_buf_init_mallocdefault(&notify_packet));
  netwib_buf_decode(&tempbuf, NETWIB_DECODETYPE_DATA, &notify_packet);
  netwib_er(netwib_buf_close(&tempbuf));
  udphdr.len = notify_packet.endoffset - notify_packet.beginoffset + 8;
  iphdr.header.ip4.totlen = udphdr.len + 20;
  ip_checksum(&iphdr); //calculate checksum for iphdr
	
  //create the complete packet to send
  netwib_er(netwib_buf_init_mallocdefault(&sendbuf)); // initiate the sendbuffer
  netwib_pkt_append_linkhdr(&linkhdr, &sendbuf);
  netwib_pkt_append_iphdr(&iphdr, &sendbuf);
  netwib_pkt_append_udphdr(&udphdr, &sendbuf);
  netwib_buf_append_buf(&notify_packet, &sendbuf);
  netwib_er(netwib_buf_close(&notify_packet));
    
  //send the packet
  ret = spoof_send(&sendbuf, &send_device);
  if (ret == NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Spoofing complete\n"));
  } 
  else {
    netwib_er(netwib_fmt_display("Failed sending SSDP:alive\n"));
  }  
  netwib_er(netwib_buf_close(&sendbuf));
  netwib_er(netwib_buf_close(&send_device));
  return(ret);
}

//Calculates the ip checksum
//input/output: *iphdr
netwib_err ip_checksum(netwib_iphdr *iphdr) {
//--------------declaration-------------------------------
  netwib_buf buf;
  netwib_err ret;
//--------------end of declaration------------------------
  
  ret = NETWIB_ERR_OK;
  netwib_er(netwib_buf_init_mallocdefault(&buf));
  netwib_pkt_append_iphdr(iphdr, &buf);
  netwib_checksum_buf(&buf, &(iphdr->header.ip4.check));
  netwib_er(netwib_buf_close(&buf));
  return(ret);
}	   

//Spoof send packet on a device
//input: *buf containing the packet, *device
netwib_err spoof_send(netwib_buf *buf, netwib_buf *device) {
//--------------declaration-------------------------------
  netwib_err ret;
  netwib_io *pio;
//--------------end of declaration------------------------
   
  ret = netwib_io_init_spoof_link(device, &pio); //initiate link spoof  
  if (ret == NETWIB_ERR_OK) {
    ret = netwib_io_write(pio, buf);
    if (ret != NETWIB_ERR_OK) {
      netwib_er(netwib_fmt_display("Sending failed\n"));
      netwib_err_display(ret, NETWIB_ERR_ENCODETYPE_FULL);
    }
    netwib_er(netwib_io_close(&pio));
  } 
  else if (ret == NETWIB_ERR_LONOTSUPPORTED) {
    netwib_er(netwib_fmt_display("IP4 spoof is not supported\n"));
  } 
  else {
    netwib_er(netwib_fmt_display("Failed initiating link spoof\n"));
    return(ret);
  }
  return(ret);   
}


//Sends a HTTP OK
//input: *serv (the service to send from), *dev (the device to send from)
//*send device (the initiated nic), *to_ip (ip to send to), *to_eth (ethernet to send to)
//to_port (port to send to)
netwib_err send_httpok(service *serv, upnp_device *dev, netwib_buf *send_device, 
           netwib_ip *to_ip, netwib_eth *to_eth, netwib_port to_port ){
           
//--------------declaration-------------------------------             
  netwib_err ret;
  netwib_time t;
  netwib_buf tempbuf, httpok_packet, sendbuf;
  netwib_iphdr iphdr;
  netwib_linkhdr linkhdr;
  netwib_udphdr udphdr;
  netwib_encodetype_context ctx;
//--------------end of declaration-------------------------------      
  
  netwib_er(netwib_fmt_display("Sending HTTP OK\n"));
  //initiating senddevice, ethhdr, iphdr and udphdr	
  netwib_linkhdr_initdefault(NETWIB_DEVICE_DLTTYPE_EN10MB, &linkhdr);
  linkhdr.hdr.ether.dst = *to_eth; //set dst mac address
  linkhdr.hdr.ether.src = dev->eth; // set src mac address
  linkhdr.hdr.ether.type = NETWIB_ETHERHDRTYPE_IP4;
  netwib_er(netwib_iphdr_initdefault(NETWIB_IPTYPE_IP4, &iphdr));
  iphdr.dst = *to_ip;
  iphdr.src = dev->ip;
  iphdr.header.ip4.id = 123;
  iphdr.protocol = NETWIB_IPPROTO_UDP;
  netwib_er(netwib_udphdr_initdefault(&udphdr));
  udphdr.dst = to_port;
  udphdr.src = serv->port;
    
  //build the HTTP OK packet
  netwib_buf_init_mallocdefault(&tempbuf);
  netwib_buf_append_text("HTTP/1.1 200 OK\r\nLOCATION: ", &tempbuf);
  netwib_buf_append_text(dev->location, &tempbuf);
  netwib_buf_append_fmt(&tempbuf, "\r\nEXT:\r\nCACHE-CONTROL: max-age = %{uint16}\r\nUSN: ",serv->timeout);
  netwib_buf_append_text(serv->usn, &tempbuf);
  netwib_buf_append_text("\r\nST: ", &tempbuf); 
  netwib_buf_append_text(serv->st, &tempbuf);
  netwib_buf_append_text("\r\nSERVER:", &tempbuf);
  netwib_buf_append_text(dev->server, &tempbuf);
  netwib_buf_append_text("\r\n\r\n", &tempbuf);
  netwib_er(netwib_buf_init_mallocdefault(&httpok_packet));
  netwib_buf_decode(&tempbuf, NETWIB_DECODETYPE_DATA, &httpok_packet);	
  udphdr.len = httpok_packet.endoffset - httpok_packet.beginoffset + 8;
  iphdr.header.ip4.totlen = udphdr.len + 20;
  ip_checksum(&iphdr); //calculate checksum for iphdr
	
  //create the complete packet to send
  netwib_er(netwib_buf_init_mallocdefault(&sendbuf)); // initiate the sendbuffer
  netwib_pkt_append_linkhdr(&linkhdr, &sendbuf);
  netwib_pkt_append_iphdr(&iphdr, &sendbuf);
  netwib_pkt_append_udphdr(&udphdr, &sendbuf);
  netwib_buf_append_buf(&httpok_packet, &sendbuf);
  netwib_er(netwib_buf_close(&httpok_packet));
	
  //send the packet
  ret = spoof_send(&sendbuf, send_device);
  if (ret == NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Spoofing complete\n"));
  } 
  else {
    netwib_er(netwib_fmt_display("Failed sending HTTP OK\n"));
  }  
  netwib_er(netwib_buf_close(&tempbuf));
  netwib_er(netwib_buf_close(&sendbuf));
  return(ret);           
}

//send arp packet
//input: *from_ip, *to_ip, *from_eth, *to_eth, *send_device
netwib_err send_arp(netwib_ip *from_ip, netwib_ip *to_ip, netwib_eth *from_eth, netwib_eth *to_eth,
                                        netwib_buf *send_device){
//--------------declaration-------------------------------             
  netwib_err ret;
  netwib_buf sendbuf;
  netwib_linkhdr linkhdr;
  netwib_arphdr arphdr;
//--------------end of declaration-------------------------------   
           
  netwib_linkhdr_initdefault(NETWIB_DEVICE_DLTTYPE_EN10MB, &linkhdr);
  linkhdr.hdr.ether.dst = *to_eth; //set dst mac address
  linkhdr.hdr.ether.src = *from_eth; // set src mac address
  //build the arp header
  linkhdr.hdr.ether.type = NETWIB_ETHERHDRTYPE_ARP;
  netwib_er(netwib_arphdr_initdefault(&arphdr));
  arphdr.ethsrc = *from_eth;
  arphdr.ipsrc = *from_ip;
  arphdr.ethdst = *to_eth;
  arphdr.ipdst = *to_ip;
  arphdr.op = NETWIB_ARPHDROP_ARPREP;
  netwib_er(netwib_buf_init_mallocdefault(&sendbuf)); // initiate the sendbuffer
  netwib_pkt_append_linkhdr(&linkhdr, &sendbuf);
  netwib_pkt_append_arphdr(&arphdr, &sendbuf);
  
  ret = spoof_send(&sendbuf, send_device);
  if (ret == NETWIB_ERR_OK) {
    netwib_er(netwib_fmt_display("Spoofing complete\n"));
  } 
  else {
    netwib_er(netwib_fmt_display("Failed sending arp\n"));
  }  
  netwib_er(netwib_buf_close(&sendbuf));
  return(ret);                          
}
           
//Get xml for a device that has Powermgmnt service and subscribes to the Powermgmnt service
//input: *dstip, *deveth, myport, *rest containing the incoming HTTP OK or SSDP alive
//       *upnp_devices (device cache), eventport, *myip
netwib_err get_xml_and_subscribe(netwib_constip *dstip, netwib_consteth *deveth, netwib_port myport,
                                netwib_buf *rest, netwib_ring *upnp_devices, netwib_port eventport, 
								netwib_constip *myip){
//--------------declaration-------------------------------
  netwib_string bufstring;
  netwib_io *pio;
  netwib_buf outpacket, inpacket, tempbuf, tempbuf2, tempbuf3;
  netwib_port dstport, subport;
  netwib_err ret;
  char port[10];
  int i;
  char url[100];
  char suburl[100];
  char xmlstring[200];
  netwib_ptr ptr;
  upnp_device *upnpdev;
  service *upnp_service;
  netwib_thread *read_events;
//--------------end of declaration------------------------

  //Parse out neede information from the incoming HTTP OK
  netwib_er(netwib_ptr_malloc(sizeof(upnp_device), &ptr));
  upnpdev = (upnp_device *) ptr;
  get_hdr(rest,"SERVER:", upnpdev->server);
  get_hdr(rest,"LOCATION:", upnpdev->location);
  netwib_buf_init_mallocdefault(&inpacket);
  netwib_buf_init_mallocdefault(&outpacket);
  //Get port and url to send HTTP Get to
  get_port_and_url(upnpdev->location, port, url, dstip);
  netwib_buf_init_mallocdefault(&tempbuf2);
  netwib_buf_init_ext_text(port, &tempbuf2); 
  netwib_port_init_buf(&tempbuf2, &dstport);
  
  //create a socket to send/receive Get for xml and subscribtion
  ret = netwib_io_init_sock_tcp_cli(myip, dstip, myport+1, dstport, &pio);
  if(ret != NETWIB_ERR_OK) {
  	netwib_fmt_display("Error while initiating socket in get_xml_and_subscribe\n");
  	return(ret);
  }
  
  //build the HTTP GET packet
  netwib_buf_init_mallocdefault(&tempbuf);
  netwib_buf_append_text("GET ", &tempbuf);
  netwib_buf_append_text(url, &tempbuf);
  netwib_buf_append_text(" HTTP/1.1\r\nHost ", &tempbuf);
  netwib_buf_append_ip(dstip, NETWIB_IP_ENCODETYPE_IP, &tempbuf);
  netwib_buf_append_text(":", &tempbuf);
  netwib_buf_append_port(dstport, &tempbuf);
  netwib_buf_append_text("\r\nContent-Lengt: 0\r\n\r\n", &tempbuf); //is this enough? Do we need language?
  netwib_buf_decode(&tempbuf, NETWIB_DECODETYPE_DATA, &outpacket);
  netwib_buf_close(&tempbuf); //closing tempbuf
  netwib_er(netwib_io_write(pio, &outpacket));
  ret = netwib_io_read(pio, &inpacket);
  netwib_io_close(&pio);
  
  //received xml for the device, create device and services
  if(ret == NETWIB_ERR_OK) {
  	//initiate new device
    upnpdev->ip = *dstip;
    upnpdev->eth = *deveth;
    netwib_er(netwib_ring_init(&delete_service_from_list, NULL, &upnpdev->services));
   	netwib_er(netwib_time_init_now(&upnpdev->last_activity));
   	ptr = (netwib_ptr *)upnpdev;
   	netwib_ring_add_first(upnp_devices, ptr);
   	
   	//add the "services" that are always sent out with ssdp:alive
   	netwib_er(netwib_ptr_malloc(sizeof(service), &ptr));
   	upnp_service = (service *) ptr;
   	upnp_service->port = dstport;
   	get_timeout(rest, &upnp_service->timeout);
   	parse_xml(&inpacket, "deviceType", xmlstring, 1);
   	netwib_c_strcpy(upnp_service->st, xmlstring);
   	parse_xml(&inpacket, "UDN", xmlstring, 1);
   	netwib_c_strcpy(upnp_service->usn, xmlstring);
   	netwib_c_strcat(upnp_service->usn, "::");
   	netwib_c_strcat(upnp_service->usn, upnp_service->st);
   	netwib_er(netwib_time_init_now(&upnp_service->last_notify)); //this is not really true, but it will work
    ptr = (netwib_ptr *)upnp_service;
    netwib_ring_add_first(upnpdev->services, ptr);
    netwib_er(netwib_ptr_malloc(sizeof(service), &ptr));
   	upnp_service = (service *) ptr;
   	upnp_service->port = dstport;
   	get_timeout(rest, &upnp_service->timeout);
   	parse_xml(&inpacket, "UDN", xmlstring, 1);
   	netwib_c_strcpy(upnp_service->st, xmlstring);
   	netwib_c_strcpy(upnp_service->usn, xmlstring);
   	netwib_er(netwib_time_init_now(&upnp_service->last_notify)); //this is not really true, but it will work
    ptr = (netwib_ptr *)upnp_service;
    netwib_ring_add_first(upnpdev->services, ptr);
    netwib_er(netwib_ptr_malloc(sizeof(service), &ptr));
   	upnp_service = (service *) ptr;
   	upnp_service->port = dstport;
   	get_timeout(rest, &upnp_service->timeout);
   	netwib_c_strcpy(upnp_service->st, "upnp:rootdevice");
   	parse_xml(&inpacket, "UDN", xmlstring, 1);
   	netwib_c_strcpy(upnp_service->usn, xmlstring);
  	netwib_c_strcat(upnp_service->usn, "::upnp:rootdevice");
   	netwib_er(netwib_time_init_now(&upnp_service->last_notify)); //this is not really true, but it will work
    ptr = (netwib_ptr *)upnp_service;
    netwib_ring_add_first(upnpdev->services, ptr);
   	
   	
   	ret = NETWIB_ERR_OK;
   	i=0;
   	//Go through all services for the device and ad to device
   	while(ret == NETWIB_ERR_OK) {
      ++i;
      ret = parse_xml(&inpacket, "serviceType", xmlstring, i);
      if (ret == NETWIB_ERR_OK) { //found another service
        if(netwib_c_strncmp(xmlstring, "urn:schemas-power:service:Powermgmt", 35) == 0) {//this is the powermgmnt service. We need subscribtion url
          //netwib_fmt_display("Found Powermgmnt service in xml\n");
		  parse_xml(&inpacket, "SCPDURL", xmlstring, i);
		  netwib__buf_reinit(&tempbuf2);
		  get_port_and_url(xmlstring, port, suburl, dstip);
		  netwib_buf_init_ext_text(port, &tempbuf2); 
          netwib_port_init_buf(&tempbuf2, &subport);
		  netwib_buf_close(&tempbuf2);//closing tempbuf2
        }
  	    netwib_er(netwib_ptr_malloc(sizeof(service), &ptr));
	  	upnp_service = (service *) ptr;
	  	upnp_service->port = dstport; 
	  	get_timeout(rest, &upnp_service->timeout);
	  	parse_xml(&inpacket, "UDN", xmlstring, 1);
   	  	netwib_c_strcpy(upnp_service->usn, xmlstring);
	  	parse_xml(&inpacket, "serviceType", xmlstring, i);
	  	netwib_c_strcpy(upnp_service->st, xmlstring);
	  	netwib_c_strcat(upnp_service->usn, "::");
	  	netwib_c_strcat(upnp_service->usn, xmlstring); 
 	  	netwib_er(netwib_time_init_now(&upnp_service->last_notify)); //this is not really true, but it will work
	  	ptr = (netwib_ptr *)upnp_service;
      	netwib_ring_add_first(upnpdev->services, ptr);
      	netwib_fmt_display("New services added to device\n");
      }
    }
    //netwib_fmt_display("New devices with service(s) added to cache, ip: %{ip}\n", dstip);
    
    //now we need to send subscribtion to powermgmnt service
    netwib_buf_init_mallocdefault(&tempbuf3);
    netwib_buf_append_text("SUBSCRIBE ", &tempbuf3);
    netwib_buf_append_text(suburl, &tempbuf3);
    netwib_buf_append_text(" HTTP/1.1\r\nHost ", &tempbuf3);
    netwib_buf_append_ip(dstip, NETWIB_IP_ENCODETYPE_IP, &tempbuf3);
    netwib_buf_append_text(":", &tempbuf3);
    netwib_buf_append_port(dstport, &tempbuf3);
    netwib_buf_append_text("\r\nCallback: <http://", &tempbuf3);
    netwib_buf_append_ip(myip, NETWIB_IP_ENCODETYPE_IP, &tempbuf3);
    netwib_buf_append_text(":", &tempbuf3);
    netwib_buf_append_port(eventport, &tempbuf3);
    netwib_buf_append_text(">\r\nNT: upnp:event\r\n", &tempbuf3);
    netwib_buf_append_text("Timeout: infinite \r\nContent-Lengt: 0\r\n\r\n", &tempbuf3);
    netwib__buf_reinit(&outpacket);
    netwib_buf_decode(&tempbuf3, NETWIB_DECODETYPE_DATA, &outpacket);
    netwib_buf_close(&tempbuf3); //closing tempbuf3
    
    ret = netwib_io_init_sock_tcp_cli(myip, dstip, myport, dstport, &pio);
    if(ret != NETWIB_ERR_OK) {
  	  netwib_fmt_display("Error while reinitiating socket in get_xml_and_subscribe\n");
  	  return(ret);
    }
   	      
    netwib_er(netwib_io_write(pio, &outpacket));
	netwib_buf_close(&outpacket); //closing outpacket
    
    ret = netwib_io_read(pio, &inpacket); //answer for our subscribtion
    netwib_io_close(&pio);
    if(ret == NETWIB_ERR_OK) {
	  netwib_er(netwib_buf_ref_string(rest, &bufstring));
	  if(netwib_c_strcasestr(bufstring, "HTTP/1.1 200 OK") == NULL) {
	  	netwib_fmt_display("Error while subscribing to power management service\n");
	  	return(NETWIB_ERR_NOTFOUND);
	  }
    }
    netwib_buf_close(&inpacket);
  }
  return(NETWIB_ERR_OK);
}

//Initiate a NIC
//input: hwt (hardware type, which kind of device is to be initiated)
//output: *pbuf (will contain description of the device), 
netwib_err init_device(netwib_buf *pbuf, netwib_device_hwtype hwt) 
{
//--------------declaration-------------------------------    
  netwib_err ret;
  netwib_conf_devices_index *pconfindex;
  netwib_conf_devices conf;
//--------------end of declaration------------------------    

  netwib_er(netwib_conf_devices_index_init(&conf, &pconfindex));
  ret = NETWIB_ERR_OK;
  while (NETWIB_TRUE) {
    ret = netwib_conf_devices_index_next(pconfindex);
    if (ret != NETWIB_ERR_OK) {
      if (ret == NETWIB_ERR_DATAEND) 
        ret = NETWIB_ERR_NOTFOUND;
      break;
    }        
    if (conf.hwtype == hwt) {
      netwib_er(netwib_buf_append_buf(&conf.device, pbuf));
      break;
    }
  }
  netwib_er(netwib_conf_devices_index_close(&pconfindex));
  if (ret == NETWIB_ERR_NOTFOUND) {
    netwib_er(netwib_fmt_display("No device of wanted type found\n"));
  }
  netwib_er(netwib_fmt_display("Device initiated\n"));
  return(ret);
}
