/*
=============================================================
proxy_updatecache.c
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the co-operating proxy program. The co-operating proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover power management services will make it possible for devices with 
these services to go into power sleep without disappearing from the network by 
sending out spoofed HTTP OK and SSDP:alive. When there is an incoming TCP 
connection for the sleeping device the proxy will wake it up using WOL.

This file contains functions used by threads to search the data structures.
The threads check if a notification for a device has expired and the devices
should be removed and if it is time to spoof SSDP:alive for a device in the
proxy cache

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program at your own risk.
=============================================================
*/

#include <stdio.h>
#include <memory.h>
#include "proxy_funcs.h"

//Thread to remove devices from chache when notification has expired
//input: infosin a *ring containing upnp_devices ring and proxy_devices ring
netwib_err update_cache(netwib_ptr infosin, netwib_ptr *pinfosout) 
{
//--------------declaration------------------------------- 
  netwib_ring_index *ri, *ri2;
  netwib_ptr ptr, ptr2, ptr3, ptr4;
  netwib_err ret, ret2;
  upnp_device *device, *device2;
  netwib_time t;
  netwib_cmp timecheck;
  netwib_ring *ring, *device_ring;
  netwib_ring *upnp_devices;
  netwib_ring *proxy_devices;
  service *upnp_service;
  netwib_ips *ips;
  netwib_uint32 nbr;
//--------------end of declaration------------------------   

  ring = (netwib_ring *)infosin;
  netwib_er(netwib_ring_index_init(ring, &ri));
  ret = netwib_ring_index_next(ri, &ptr);
  proxy_devices = (netwib_ring *)ptr;
  ret = netwib_ring_index_next(ri, &ptr);
  upnp_devices = (netwib_ring *)ptr;
  ret = netwib_ring_index_next(ri, &ptr);
  ips = (netwib_ips *)ptr;
    
  //infinite loop to update proxy_devices
  while(NETWIB_TRUE) {
    ret = NETWIB_ERR_OK;
    netwib_er(netwib_ring_index_init(upnp_devices, &ri));
    if(ret != NETWIB_ERR_OK){
       netwib_fmt_display("Error while initiating index in update cache.");
       return(NETWIB_ERR_NOTFOUND);
    }
    ret = netwib_ring_index_next(ri, &ptr);
    //loop to go through all devices every time
    while(ret == NETWIB_ERR_OK) 
	{
      device = (upnp_device *)ptr;
      device_ring = device->services;
      netwib_er(netwib_ring_index_init(device_ring, &ri2));
      ret2 = netwib_ring_index_next(ri2, &ptr);
      upnp_service = (service *)ptr;
      //find the power mgmt service 
      while(ret2 != NETWIB_ERR_DATAEND && strncmp("urn:schemas-power:service:Powermgmt", upnp_service->st, 35)!= 0){
        ret2 = netwib_ring_index_next(ri2, &ptr);
        upnp_service = (service *)ptr;                  
      }
      netwib_er(netwib_ring_index_close(&ri2));
      //after exiting this loop we must have found the powermgmt
      netwib_er(netwib_time_init_now(&t));
	  netwib_er(netwib_time_minus_sec(&t, upnp_service->timeout)); //check for the timeout 
	  netwib_time_cmp(&upnp_service->last_notify, &t, &timecheck); //check if the last notify has expired
	  if( timecheck == NETWIB_CMP_LT ) {//notification has expired remove from proxy and cache
	    netwib_fmt_display("Notification expired, removing %{ip} from cache\n", &device->ip);
	    ret = remove_device(&device->ip, proxy_devices);
	    ret = remove_device(&device->ip, upnp_devices);
	    if(ret != NETWIB_ERR_OK)
          netwib_fmt_display("Error when removing device from cache");
	    ret = netwib_ips_del_ip(ips, &device->ip);
	  }
	  netwib_ring_ctl_get_count(upnp_devices,&nbr);
	  if(nbr == 0) {
	    ret = NETWIB_ERR_DATAEND;
	  }
	  else {
	    ret = netwib_ring_index_next(ri, &ptr);
	  }
    }
    netwib_er(netwib_ring_index_close(&ri));
    netwib_time_sleep_sec(5); //sleep 5 sec
  } 
}

//Copies a device from the device cache to the proxy cache
//input: *ip (ip of device to copy), *cache (device cache), *proxy (proxy cache)
netwib_err copy_from_cache_to_proxy(netwib_ip *ip, netwib_ring *cache, netwib_ring *proxy){
//--------------declaration------------------------------- 
  upnp_device *device, *device2;
  netwib_ptr ptr, ptr2, ptr3, ptr4;
  netwib_ring_index *ri;
  netwib_err ret;
//--------------end of declaration------------------------
  
  find_ip(ip, cache, &ptr); //find device to copy
  device = (upnp_device *)ptr;
  netwib_er(netwib_ptr_malloc(sizeof(upnp_device), &ptr2));
  //copy the device to proxy_devices
  netwib_c_memcpy(ptr2, ptr, sizeof(upnp_device));
  device2 = (upnp_device *)ptr2;
  //initiate a new ring
  netwib_er(netwib_ring_init(&delete_service_from_list, NULL, &device2->services));
  netwib_er(netwib_ring_index_init(device->services, &ri));
  //copy every item from the first ring to the second
  ret = netwib_ring_index_next(ri, &ptr3);
  while(ret != NETWIB_ERR_DATAEND){                        
    netwib_er(netwib_ptr_malloc(sizeof(service), &ptr4));
    netwib_c_memcpy(ptr4, ptr3, sizeof(service));
    netwib_ring_add_first(device2->services, ptr4);
    ret = netwib_ring_index_next(ri, &ptr3);
  }		  	 
  netwib_er(netwib_ring_add_first(proxy, ptr2));// add device to proxy list         
}

//thread to send out SSDP:alive whenever needed
//input infosin a *ring containing upnp_devices ring and proxy_devices ring
//will send out SSDP:alive for devieces in proxymode after 1/3 of the timout time since last SSDP:alive was sent
netwib_err spoof_notify(netwib_ptr infosin, netwib_ptr *pinfosout)
{
//--------------declaration------------------------------- 
  netwib_ring_index *ri;
  netwib_ring_index *rid; 
  netwib_ring_index *ris;
  netwib_err ret1, ret2;
  netwib_time t;
  netwib_ptr ptr1;
  netwib_ptr ptr2;
  netwib_ptr deviceptr, servptr;
  upnp_device *device, *device2;
  service *serv;
  netwib_cmp timecheck;
  netwib_ring *ring;
  netwib_ring *upnp_devices;
  netwib_ring *proxy_devices;
//--------------end of declaration------------------------ 

  ring = (netwib_ring *)infosin;
  netwib_er(netwib_ring_index_init(ring, &ri));
  ret1 = netwib_ring_index_next(ri, &ptr1);
  proxy_devices = (netwib_ring *)ptr1;
  ret1 = netwib_ring_index_next(ri, &ptr1);
  upnp_devices = (netwib_ring *)ptr1;
  netwib_er(netwib_ring_index_close(&ri));
  // inifinite loop to send SSDP:alive 
  while(NETWIB_TRUE) {
    ret1 = NETWIB_ERR_OK;
    netwib_er(netwib_ring_index_init(proxy_devices, &rid));
    ret1 = netwib_ring_index_next(rid, &ptr1);
    //loop to go through all devices in proxy_devices
    while(ret1 != NETWIB_ERR_DATAEND) {
      //netwib_fmt_display("found a device in proxy list\n");
	  device = (upnp_device *)ptr1;
	  ret2 = NETWIB_ERR_OK;
	  netwib_er(netwib_ring_index_init(device->services, &ris));
	  ret1 = netwib_ring_index_next(ris, &ptr2);
	  //loop to go through all services of a device
	  while(ret2 != NETWIB_ERR_DATAEND) {
   	    serv = (service *) ptr2;
		netwib_er(netwib_time_init_now(&t));
		netwib_er(netwib_time_minus_sec(&t, (serv->timeout/3)));
		netwib_time_cmp(&(serv->last_notify), &t, &timecheck);
		//it is time to send a SSDP:alive
		if(timecheck == NETWIB_CMP_LT) {
          ret1 = send_alive(serv, device);// send SSDP:alive for service and update last_notify
          if(ret1 == NETWIB_ERR_OK) {
		    netwib_er(netwib_time_init_now(&serv->last_notify));
		    ret1 = find_ip(&device->ip, upnp_devices, &deviceptr);
		    if(ret1 == NETWIB_ERR_OK) {
		    device2 = (upnp_device *) deviceptr;
		      ret1 = find_usn(serv->usn, device2->services, &servptr);
		      if(ret1 == NETWIB_ERR_OK) {
		        serv = (service *) servptr;
		      	netwib_er(netwib_time_init_now(&serv->last_notify));
		      }	
            }
		  }  
		}
		ret2 = netwib_ring_index_next(ris, &ptr2);
	  }
	  netwib_er(netwib_ring_index_close(&ris));
	  ret1 = netwib_ring_index_next(rid, &ptr1);
	}
	
	netwib_er(netwib_ring_index_close(&rid));
  	netwib_time_sleep_sec(5);	//sleep for 5 seconds  
  }	   
}
